/*
* Copyright (c) 2013, 2016, Oracle and/or its affiliates. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */


/*
 * @test
 * @bug 8185712
 * @summary Callstacks shall be printed correctly after a sigsegv.
 * @library /test/lib
 * @requires (vm.debug == true) & (os.family == "windows")
 * @author Thomas Stuefe (SAP)
 * @modules java.base/jdk.internal.misc
 *          java.management
 */

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.regex.Pattern;

import jdk.test.lib.process.OutputAnalyzer;
import jdk.test.lib.Platform;
import jdk.test.lib.process.ProcessTools;

public class CallstackAfterSegfault {

  // Whether callstack should show the name of the binary (shared lib/executable)
  final static boolean with_binary_name = true;
  // Whether callstack should show File information (source file, line)
  final static boolean with_source_file_info = Platform.isWindows();

  private static String buildFramePattern(String library, String functionName, String sourcefile) {
    StringBuilder bld = new StringBuilder();
    bld.append(".*");
    if (with_binary_name) {
        // We only append the raw library name without extensions or prefix, to match on all platforms.
        bld.append(library);
        bld.append(".*");
    }
    bld.append(functionName);
    bld.append(".*");
    if (with_source_file_info) {
        bld.append(sourcefile);
        bld.append(":\\d+");
        bld.append(".*");
    }
    return bld.toString();
  }

  public static void main(String[] args) throws Exception {

    ProcessBuilder pb = ProcessTools.createJavaProcessBuilder(
        "-XX:+UnlockDiagnosticVMOptions",
        "-Xmx32M",
        "-XX:-CreateCoredumpOnCrash",
        "-XX:ErrorHandlerTest=14",
        "-version");

    OutputAnalyzer output_detail = new OutputAnalyzer(pb.start());

    // we should have crashed with a SIGFPE
    output_detail.shouldMatch("# A fatal error has been detected by the Java Runtime Environment:.*");
    output_detail.shouldMatch("# +EXCEPTION_ACCESS_VIOLATION.*");

    // extract hs-err file
    String hs_err_file = output_detail.firstMatch("# *(\\S*hs_err_pid\\d+\\.log)", 1);
    if (hs_err_file == null) {
      throw new RuntimeException("Did not find hs-err file in output.\n");
    }

    // scan hs-err file: File should contain the "[error occurred during error reporting..]"
    // markers which show that the secondary error handling kicked in and handled the
    // error successfully. As an added test, we check that the last line contains "END.",
    // which is an end marker written in the last step and proves that hs-err file was
    // completely written.
    File f = new File(hs_err_file);
    if (!f.exists()) {
      throw new RuntimeException("hs-err file missing at "
          + f.getAbsolutePath() + ".\n");
    }

    System.out.println("Found hs_err file at " +  f.getAbsolutePath() + ". Scanning...");

    FileInputStream fis = new FileInputStream(f);
    BufferedReader br = new BufferedReader(new InputStreamReader(fis));
    String line = null;

    Pattern [] pattern = new Pattern[] {
        /* Note: some platforms (e.g. AIX) have demangling off by default. Formulate the pattern such
         * that function names match for both mangled and unmangled case. */
        Pattern.compile(buildFramePattern("jvm", "crash_with_segfault", "vmerror.cpp")),
        Pattern.compile(buildFramePattern("jvm", "controlled_crash", "vmerror.cpp")),
        Pattern.compile(buildFramePattern("jvm", "JNI_CreateJavaVM", "jni.cpp"))
    };

    // We require at least one of the above pattern to match.

    int numMatches = 0;
    while ((line = br.readLine()) != null) {
      for (int i = 0; i < pattern.length; i ++) {
        if (pattern[i].matcher(line).matches()) {
          System.out.println("Found: " + line + ".");
          numMatches ++;
          break;
        }
      }
      if (numMatches == pattern.length) {
        break;
      }
    }
    br.close();

    if (numMatches == 0) {
      throw new RuntimeException("Missing stack frames in hs-err file.");
    }

    System.out.println("OK.");

  }

}


